import os
import logging
import pathlib

import io
from typing import Optional, Callable

from pytube import YouTube, StreamQuery, Stream
from pytube.exceptions import PytubeError

from suskabot.utils import utils


logger: logging.Logger = logging.getLogger()


class VideoTooLarge(Exception):
    """Raise if a video file size exceeds set limits"""


class YTVideoMetadata:
    title: str
    filesize_bytes: int

    def __init__(
            self,
            title: str,
            filesize_bytes: int,
    ) -> None:
        self.title = title
        self.filesize_bytes = filesize_bytes


class YTDownloaderService:
    _minimum_video_resolution: int
    _maximum_video_resolution: int
    _file_size_limit_mb: int
    _save_to_drive: bool

    _default_download_path = pathlib.Path(os.getcwd()).joinpath("yt_downloads/")

    def __init__(
            self,
            min_video_res: int,
            max_video_res: int,
            file_size_limit_mb: int,
            save_to_drive: bool
    ) -> None:
        self._minimum_video_resolution = min_video_res
        self._maximum_video_resolution = max_video_res
        self._file_size_limit_mb = file_size_limit_mb
        self._save_to_drive = save_to_drive

    def _get_best_quality_stream_under_limits(self, video_streams: StreamQuery) -> Optional[Stream]:
        ordered_video_streams: StreamQuery = video_streams.order_by("resolution").desc()
        logger.debug(f"Trying to find a stream under {self._file_size_limit_mb} MB, between "
                     f"{self._minimum_video_resolution}p and {self._maximum_video_resolution}p")

        video_stream: Stream
        for video_stream in ordered_video_streams:
            video_resolution: int = utils.pytube_video_resolution_to_int(video_stream.resolution)
            video_filesize_mb: int = utils.bytes_to_megabytes(video_stream.filesize)
            logger.debug(f"Checking a stream {video_stream.itag} with {video_stream.resolution} resolution "
                         f"and {video_filesize_mb} MB filesize")
            # noinspection PyChainedComparisons
            # I don't like how chained comparisons look.
            if (video_filesize_mb <= self._file_size_limit_mb and
                    video_resolution >= self._minimum_video_resolution and
                    video_resolution <= self._maximum_video_resolution):
                logger.debug(f"Selected stream {video_stream.itag}")
                return video_stream
        else:
            logger.debug("No video stream fits the requirements")
            return None

    # all the stream availability exceptions are raised here!
    def get_youtube_video_stream(self, video_url: str) -> Optional[Stream]:
        """
        Gets YouTube video stream under set limits from a valid video url.

        :param video_url: valid YouTube video url

        :return: pytube Stream object, right now only progressive streams are supported

        :raises RegexMatchError: If video url is invalid
        :raises VideoUnavailable: If the video is unavailable for some reason
        :raises VideoTooLarge: If video file size exceeds set limits
        """

        try:
            video_data = YouTube(url=video_url)
            video_data.check_availability()
        except PytubeError as e:
            logger.exception("Failed to get video stream data: ", exc_info=e)
            raise e

        progressive_video_streams: StreamQuery = video_data.streams.filter(progressive=True)
        logger.debug(f"Found progressive streams:\n{progressive_video_streams}")

        video_stream: Stream | None = self._get_best_quality_stream_under_limits(progressive_video_streams)
        logger.info(f"Selected video stream:\n{video_stream}")

        if not video_stream:
            raise VideoTooLarge(f"Can't download videos larger than {self._file_size_limit_mb}MB")

        return video_stream

    def download_youtube_video_stream(
            self,
            video_stream: Stream,
            progress_callback: Callable[[Stream, bytes, int], None] | None = None
    ) -> bytes | None:
        """
        Download YouTube video from given stream and get bytes of it!
        :param video_stream: pytube Stream object, needs to be obtained from get_youtube_video_stream
        :param progress_callback: optional on_progress callback,
        needs to have these three parameters:
        1) Stream, video stream that is being downloaded
        2) bytes, chunk of bytes downloaded
        3) int, number of bytes remaining

        :return: Bytes of the downloaded video or none

        :raises OSError: If save_to_drive is enabled and output file couldn't be opened
        """
        # the only way to register a callback is at the moment of YouTube object creation.
        # since this function only has a stream object to work with,
        # registering a callback is impossible by normal means.
        # one more reason to migrate from pytube, lol
        video_stream._monostate.on_progress = progress_callback

        if self._save_to_drive:
            video_path: str = video_stream.download(
                output_path=str(self._default_download_path)
            )
        else:
            with io.BytesIO() as buffer:
                video_stream.stream_to_buffer(buffer)
                logger.info(f"Successfully downloaded {video_stream.title} to RAM")
                return buffer.getvalue()

        with open(video_path, "rb") as video_file:
            logger.info(f"Successfully downloaded {video_stream.url} to drive")
            return video_file.read()

    # doesn't serve any purpose as of now, but I'll leave it here for a bit
    # def __clear_download_directory(self) -> None:
    #     try:
    #         shutil.rmtree(self._default_download_path)
    #     except Exception as e:
    #         logger.warning(f"Could not clear download directory, {e}")
    #         raise e



from telegram.ext import ApplicationBuilder, Application

from suskabot.config.config import Config
from suskabot.controllers.telegram_api.bot import TGBot
from suskabot.services.translator import TranslatorService
from suskabot.services.youtube_downloader import YTDownloaderService


class App:
    _translator_service: TranslatorService
    _youtube_downloader_service: YTDownloaderService

    _ptb_app: Application
    _telegram_bot: TGBot

    def __init__(
            self,
            config: Config
    ) -> None:
        # initialize services
        self._translator_service = TranslatorService(
            comma_separated_translation_services=config.comma_separated_translation_services,
            user_language=config.user_language,
            default_lang_to_translate_to=config.default_language_to_translate_to
        )

        self._youtube_downloader_service = YTDownloaderService(
            file_size_limit_mb=config.file_size_limit_mb,
            min_video_res=config.minimum_video_resolution,
            max_video_res=config.maximum_video_resolution,
            save_to_drive=config.save_to_drive
        )

        # initialize controllers
        self._ptb_app = ApplicationBuilder().token(config.tg_bot_api_token).build()
        self._telegram_bot = TGBot(
            application=self._ptb_app,
            translator_service=self._translator_service,
            yt_downloader_service=self._youtube_downloader_service
        )

    def run(self) -> None:
        self._telegram_bot.start()



import logging
import os
import enum
from typing import Self, Dict, Optional

import telegram.constants
import pydantic
import dotenv

from suskabot.utils import utils


logger = logging.getLogger()


class VideoResolution(enum.Enum):
    RES_144 = 144
    RES_240 = 240
    RES_360 = 360
    RES_480 = 480
    RES_720 = 720
    RES_1080 = 1080
    RES_1440 = 1440
    RES_2160 = 2160


def get_lowest_resolution_value() -> int:
    return VideoResolution.RES_144.value


def get_highest_resolution_value() -> int:
    return VideoResolution.RES_2160.value


# it is assumed that every valid configuration is type annotated
# I tried really hard, but there's no easier way to achieve the same functionality without repeating config names
def get_config_attributes() -> set[str]:
    return vars(Config)['__annotations__'].keys()


# every configuration must be type annotated, else it won't be loaded!!!
class Config(pydantic.BaseModel):
    tg_bot_api_token: str

    comma_separated_translation_services: str
    user_language: str
    default_language_to_translate_to: str

    minimum_video_resolution: int
    maximum_video_resolution: int
    file_size_limit_mb: int
    save_to_drive: bool

    tg_bot_api_token = pydantic.Field(
        pattern=r"^[0-9]+:[a-zA-Z0-9_-]+$"
    )

    comma_separated_translation_services = pydantic.Field(
        default="google,yandex,bing,argos",
        pattern=r"^([a-zA-Z]+,?)+"
    )
    user_language = pydantic.Field(
        default="ru",
        pattern=r"^[^-]{2,3}(-[^-]{2,3})?(-[^-]{2,3})?$"
    )
    default_language_to_translate_to = pydantic.Field(
        default="en",
        pattern=r"^[^-]{2,3}(-[^-]{2,3})?(-[^-]{2,3})?$",
    )

    # right now only progressive videos are supported, so video quality is 720p max.
    # it also looks like 360p and 720p are the only available options
    minimum_video_resolution = pydantic.Field(
        default=get_lowest_resolution_value(),
    )
    maximum_video_resolution = pydantic.Field(
        default=get_highest_resolution_value(),
    )

    @pydantic.field_validator("maximum_video_resolution", "minimum_video_resolution")
    def validate_resolution(cls, v):
        valid_video_resolutions = [res.value for res in VideoResolution]
        assert v in valid_video_resolutions, ("this does not look like a valid resolution. "
                                              f"supported resolutions: {valid_video_resolutions}")
        return v

    @pydantic.model_validator(mode='after')
    def check_if_valid_video_resolution_preferences(self) -> Self:
        if self.minimum_video_resolution > self.maximum_video_resolution:
            raise ValueError("minimum video resolution can't be higher than maximum video resolution")
        if self.minimum_video_resolution > 720:
            logger.warning("Since only progressive videos are supported right now,"
                           " 720p is the maximum possible resolution")
        return self

    # setting the limit higher than the default will break telegram api controller!
    file_size_limit_mb = pydantic.Field(
        default=utils.bytes_to_megabytes(
            telegram.constants.FileSizeLimit.FILESIZE_UPLOAD.value
        ),
        gt=0,
        validate_default=True
    )

    save_to_drive = pydantic.Field(
        default=True
    )

    # TODO: having a config for yt video downloads directory seems like a nice idea


def load_config() -> Optional[Config]:
    """
    Loads configurations from os.environ first and /project_name/config/.env file second,
    then validates them using pydantic

    :return: Config instance or None

    :raises pydantic.ValidationError: If some configurations are invalid
    """
    env_file_config: Dict[str, str | None] = dotenv.dotenv_values("./suskabot/config/.env")

    config_data: Dict[str, str | None] = {}

    configurations: set[str] = get_config_attributes()
    configuration: str
    for configuration in configurations:
        configuration_value: str | None
        configuration_upper: str = configuration.upper()

        logger.debug(f"Getting {configuration}..")

        if configuration_upper in os.environ and os.environ.get(configuration_upper) is not None:
            configuration_value = os.environ[configuration_upper]
            logger.debug(f"Got {configuration} from environ")

        elif configuration_upper in env_file_config and env_file_config.get(configuration_upper) != "":
            configuration_value = env_file_config[configuration_upper]
            logger.debug(f"Got {configuration} from .env")

        else:
            logger.debug(f"Couldn't find {configuration} anywhere, will try to use the default value")
            continue

        config_data[configuration] = configuration_value
        logger.info(f"{configuration} set to {configuration_value}")

    config: Config = Config(**config_data)

    return config



__all__ = [
    'app',
    'config',
    'controllers',
    'services',
    'utils'
]

import logging

logging.basicConfig(
    format='[%(asctime)s] '
           'Thread: %(threadName)s, '
           'module: %(module)s '
           'at %(funcName)s '
           '(line %(lineno)s) '
           '- %(levelname)s - '
           '%(message)s',
    level=logging.INFO
)
if logging.root.level > logging.DEBUG:
    # prevents log spam from infinite polling
    logging.getLogger("httpx").setLevel(logging.WARNING)



import logging
import sys

from typing import Optional
from contextlib import suppress

from suskabot.config import config

from suskabot.app import app


logger = logging.getLogger()


# somehow pydantic raises an attribute error when I'm trying to access config.Config for the type hint
# removing the type hint solves the issue, but I'd rather just suppress the pydantic error
with suppress(AttributeError):
    config: Optional[config.Config] = config.load_config()
if not config:
    logger.error("Couldn't load config file")
    sys.exit(1)

app = app.App(config)

app.run()
